/*******************************************************************************
 *     ___                  _   ____  ____
 *    / _ \ _   _  ___  ___| |_|  _ \| __ )
 *   | | | | | | |/ _ \/ __| __| | | |  _ \
 *   | |_| | |_| |  __/\__ \ |_| |_| | |_) |
 *    \__\_\\__,_|\___||___/\__|____/|____/
 *
 *  Copyright (c) 2014-2019 Appsicle
 *  Copyright (c) 2019-2022 QuestDB
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 ******************************************************************************/

package io.questdb.griffin.engine.functions;

import io.questdb.cairo.ColumnType;
import io.questdb.cairo.sql.Record;
import io.questdb.cairo.sql.RecordCursorFactory;
import io.questdb.cairo.sql.ScalarFunction;
import io.questdb.std.BinarySequence;
import io.questdb.std.Long256;
import io.questdb.std.str.CharSink;

public abstract class BooleanFunction implements ScalarFunction {

    @Override
    public final BinarySequence getBin(Record rec) {
        throw new UnsupportedOperationException();
    }

    @Override
    public long getBinLen(Record rec) {
        throw new UnsupportedOperationException();
    }

    @Override
    public final byte getByte(Record rec) {
        return (byte) (getBool(rec) ? 0 : 1);
    }

    @Override
    public final char getChar(Record rec) {
        return getBool(rec) ? 'T' : 'F';
    }

    @Override
    public final long getDate(Record rec) {
        return getBool(rec) ? 0 : 1;
    }

    @Override
    public final double getDouble(Record rec) {
        return getBool(rec) ? 0 : 1;
    }

    @Override
    public final float getFloat(Record rec) {
        return getBool(rec) ? 0 : 1;
    }

    @Override
    public final int getInt(Record rec) {
        return getBool(rec) ? 0 : 1;
    }

    @Override
    public final long getLong(Record rec) {
        return getBool(rec) ? 0 : 1;
    }

    @Override
    public final void getLong256(Record rec, CharSink sink) {
        throw new UnsupportedOperationException();
    }

    @Override
    public final Long256 getLong256A(Record rec) {
        throw new UnsupportedOperationException();
    }

    @Override
    public final Long256 getLong256B(Record rec) {
        throw new UnsupportedOperationException();
    }

    @Override
    public final RecordCursorFactory getRecordCursorFactory() {
        throw new UnsupportedOperationException();
    }

    @Override
    public final short getShort(Record rec) {
        return (short) (getBool(rec) ? 0 : 1);
    }

    @Override
    public final CharSequence getStr(Record rec) {
        return getStr0(rec);
    }

    @Override
    public final void getStr(Record rec, CharSink sink) {
        sink.put(getStr0(rec));
    }

    @Override
    public final CharSequence getStrB(Record rec) {
        return getStr0(rec);
    }

    @Override
    public final int getStrLen(Record rec) {
        return getBool(rec) ? "true".length() : "false".length();
    }

    @Override
    public final CharSequence getSymbol(Record rec) {
        return getStr0(rec);
    }

    @Override
    public final CharSequence getSymbolB(Record rec) {
        return getStr0(rec);
    }

    @Override
    public long getTimestamp(Record rec) {
        return getBool(rec) ? 0 : 1;
    }

    @Override
    public byte getGeoByte(Record rec) {
        throw new UnsupportedOperationException();
    }

    @Override
    public short getGeoShort(Record rec) {
        throw new UnsupportedOperationException();
    }

    @Override
    public int getGeoInt(Record rec) {
        throw new UnsupportedOperationException();
    }

    @Override
    public long getGeoLong(Record rec) {
        throw new UnsupportedOperationException();
    }

    @Override
    public final int getType() {
        return ColumnType.BOOLEAN;
    }

    private String getStr0(Record rec) {
        return getBool(rec) ? "true" : "false";
    }
}
